@@ -11,15 +11,15 @@
 extern crate uucore;
 
 use clap::{crate_version, App, Arg};
-use std::cmp::{min, Ordering};
+use std::cmp::Ordering;
 use std::fs::File;
 use std::io::{stdin, BufRead, BufReader, Lines, Stdin};
+use uucore::display::Quotable;
 
 static NAME: &str = "join";
 
 #[derive(Copy, Clone, PartialEq)]
 enum FileNum {
-    None,
     File1,
     File2,
 }
@@ -41,7 +41,8 @@ enum CheckOrder {
 struct Settings {
     key1: usize,
     key2: usize,
-    print_unpaired: FileNum,
+    print_unpaired1: bool,
+    print_unpaired2: bool,
     print_joined: bool,
     ignore_case: bool,
     separator: Sep,
@@ -57,7 +58,8 @@ impl Default for Settings {
         Settings {
             key1: 0,
             key2: 0,
-            print_unpaired: FileNum::None,
+            print_unpaired1: false,
+            print_unpaired2: false,
             print_joined: true,
             ignore_case: false,
             separator: Sep::Whitespaces,
@@ -101,17 +103,12 @@ impl<'a> Repr<'a> {
     }
 
     /// Print each field except the one at the index.
-    fn print_fields(&self, line: &Line, index: usize, max_fields: Option<usize>) {
-        for i in 0..min(max_fields.unwrap_or(usize::max_value()), line.fields.len()) {
+    fn print_fields(&self, line: &Line, index: usize) {
+        for i in 0..line.fields.len() {
             if i != index {
                 print!("{}{}", self.separator, line.fields[i]);
             }
         }
-        if let Some(n) = max_fields {
-            for _ in line.fields.len()..n {
-                print!("{}", self.separator)
-            }
-        }
     }
 
     /// Print each field or the empty filler if the field is not set.
@@ -185,18 +182,18 @@ impl Spec {
                     return Spec::Key;
                 }
 
-                crash!(1, "invalid field specifier: '{}'", format);
+                crash!(1, "invalid field specifier: {}", format.quote());
             }
             Some('1') => FileNum::File1,
             Some('2') => FileNum::File2,
-            _ => crash!(1, "invalid file number in field spec: '{}'", format),
+            _ => crash!(1, "invalid file number in field spec: {}", format.quote()),
         };
 
         if let Some('.') = chars.next() {
             return Spec::Field(file_num, parse_field_number(chars.as_str()));
         }
 
-        crash!(1, "invalid field specifier: '{}'", format);
+        crash!(1, "invalid field specifier: {}", format.quote());
     }
 }
 
@@ -232,7 +229,6 @@ struct State<'a> {
     print_unpaired: bool,
     lines: Lines<Box<dyn BufRead + 'a>>,
     seq: Vec<Line>,
-    max_fields: Option<usize>,
     line_num: usize,
     has_failed: bool,
 }
@@ -243,14 +239,14 @@ impl<'a> State<'a> {
         name: &'a str,
         stdin: &'a Stdin,
         key: usize,
-        print_unpaired: FileNum,
+        print_unpaired: bool,
     ) -> State<'a> {
         let f = if name == "-" {
             Box::new(stdin.lock()) as Box<dyn BufRead>
         } else {
             match File::open(name) {
                 Ok(file) => Box::new(BufReader::new(file)) as Box<dyn BufRead>,
-                Err(err) => crash!(1, "{}: {}", name, err),
+                Err(err) => crash!(1, "{}: {}", name.maybe_quote(), err),
             }
         };
 
@@ -258,10 +254,9 @@ impl<'a> State<'a> {
             key,
             file_name: name,
             file_num,
-            print_unpaired: print_unpaired == file_num,
+            print_unpaired,
             lines: f.lines(),
             seq: Vec::new(),
-            max_fields: None,
             line_num: 0,
             has_failed: false,
         }
@@ -328,8 +323,8 @@ impl<'a> State<'a> {
                     });
                 } else {
                     repr.print_field(key);
-                    repr.print_fields(line1, self.key, self.max_fields);
-                    repr.print_fields(line2, other.key, other.max_fields);
+                    repr.print_fields(line1, self.key);
+                    repr.print_fields(line2, other.key);
                 }
 
                 println!();
@@ -360,14 +355,15 @@ impl<'a> State<'a> {
         !self.seq.is_empty()
     }
 
-    fn initialize(&mut self, read_sep: Sep, autoformat: bool) {
+    fn initialize(&mut self, read_sep: Sep, autoformat: bool) -> usize {
         if let Some(line) = self.read_line(read_sep) {
-            if autoformat {
-                self.max_fields = Some(line.fields.len());
-            }
-
             self.seq.push(line);
+
+            if autoformat {
+                return self.seq[0].fields.len();
+            }
         }
+        0
     }
 
     fn finalize(&mut self, input: &Input, repr: &Repr) {
@@ -398,7 +394,11 @@ impl<'a> State<'a> {
         let diff = input.compare(self.get_current_key(), line.get_field(self.key));
 
         if diff == Ordering::Greater {
-            eprintln!("{}:{}: is not sorted", self.file_name, self.line_num);
+            eprintln!(
+                "{}:{}: is not sorted",
+                self.file_name.maybe_quote(),
+                self.line_num
+            );
 
             // This is fatal if the check is enabled.
             if input.check_order == CheckOrder::Enabled {
@@ -430,7 +430,7 @@ impl<'a> State<'a> {
             });
         } else {
             repr.print_field(line.get_field(self.key));
-            repr.print_fields(line, self.key, self.max_fields);
+            repr.print_fields(line, self.key);
         }
 
         println!();
@@ -450,11 +450,19 @@ pub fn uumain(args: impl uucore::Args) -> i32 {
 
     let mut settings: Settings = Default::default();
 
-    if let Some(value) = matches.value_of("v") {
-        settings.print_unpaired = parse_file_number(value);
+    let v_values = matches.values_of("v");
+    if v_values.is_some() {
         settings.print_joined = false;
-    } else if let Some(value) = matches.value_of("a") {
-        settings.print_unpaired = parse_file_number(value);
+    }
+
+    let unpaired = v_values
+        .unwrap_or_default()
+        .chain(matches.values_of("a").unwrap_or_default());
+    for file_num in unpaired {
+        match parse_file_number(file_num) {
+            FileNum::File1 => settings.print_unpaired1 = true,
+            FileNum::File2 => settings.print_unpaired2 = true,
+        }
     }
 
     settings.ignore_case = matches.is_present("i");
@@ -503,7 +511,7 @@ pub fn uumain(args: impl uucore::Args) -> i32 {
         crash!(1, "both files cannot be standard input");
     }
 
-    exec(file1, file2, &settings)
+    exec(file1, file2, settings)
 }
 
 pub fn uu_app() -> App<'static, 'static> {
@@ -520,7 +528,8 @@ When FILE1 or FILE2 (not both) is -, read standard input.",
         .arg(
             Arg::with_name("a")
                 .short("a")
-                .takes_value(true)
+                .multiple(true)
+                .number_of_values(1)
                 .possible_values(&["1", "2"])
                 .value_name("FILENUM")
                 .help(
@@ -531,6 +540,9 @@ FILENUM is 1 or 2, corresponding to FILE1 or FILE2",
         .arg(
             Arg::with_name("v")
                 .short("v")
+                .multiple(true)
+                .number_of_values(1)
+                .possible_values(&["1", "2"])
                 .value_name("FILENUM")
                 .help("like -a FILENUM, but suppress joined output lines"),
         )
@@ -609,7 +621,7 @@ FILENUM is 1 or 2, corresponding to FILE1 or FILE2",
         )
 }
 
-fn exec(file1: &str, file2: &str, settings: &Settings) -> i32 {
+fn exec(file1: &str, file2: &str, settings: Settings) -> i32 {
     let stdin = stdin();
 
     let mut state1 = State::new(
@@ -617,7 +629,7 @@ fn exec(file1: &str, file2: &str, settings: &Settings) -> i32 {
         file1,
         &stdin,
         settings.key1,
-        settings.print_unpaired,
+        settings.print_unpaired1,
     );
 
     let mut state2 = State::new(
@@ -625,7 +637,7 @@ fn exec(file1: &str, file2: &str, settings: &Settings) -> i32 {
         file2,
         &stdin,
         settings.key2,
-        settings.print_unpaired,
+        settings.print_unpaired2,
     );
 
     let input = Input::new(
@@ -634,18 +646,34 @@ fn exec(file1: &str, file2: &str, settings: &Settings) -> i32 {
         settings.check_order,
     );
 
+    let format = if settings.autoformat {
+        let mut format = vec![Spec::Key];
+        let mut initialize = |state: &mut State| {
+            let max_fields = state.initialize(settings.separator, settings.autoformat);
+            for i in 0..max_fields {
+                if i != state.key {
+                    format.push(Spec::Field(state.file_num, i));
+                }
+            }
+        };
+        initialize(&mut state1);
+        initialize(&mut state2);
+        format
+    } else {
+        state1.initialize(settings.separator, settings.autoformat);
+        state2.initialize(settings.separator, settings.autoformat);
+        settings.format
+    };
+
     let repr = Repr::new(
         match settings.separator {
             Sep::Char(sep) => sep,
             _ => ' ',
         },
-        &settings.format,
+        &format,
         &settings.empty,
     );
 
-    state1.initialize(settings.separator, settings.autoformat);
-    state2.initialize(settings.separator, settings.autoformat);
-
     if settings.headers {
         state1.print_headers(&state2, &repr);
         state1.reset_read_line(&input);
@@ -704,7 +732,7 @@ fn get_field_number(keys: Option<usize>, key: Option<usize>) -> usize {
 fn parse_field_number(value: &str) -> usize {
     match value.parse::<usize>() {
         Ok(result) if result > 0 => result - 1,
-        _ => crash!(1, "invalid field number: '{}'", value),
+        _ => crash!(1, "invalid field number: {}", value.quote()),
     }
 }
 
@@ -712,7 +740,7 @@ fn parse_file_number(value: &str) -> FileNum {
     match value {
         "1" => FileNum::File1,
         "2" => FileNum::File2,
-        value => crash!(1, "invalid file number: '{}'", value),
+        value => crash!(1, "invalid file number: {}", value.quote()),
     }
 }
 
